@@ -54,6 +54,7 @@
     backing: (NSBackingStoreType) bufferingType
     defer: (BOOL) flag screen: (NSScreen *) aScreen
     gstWin: (GstGLWindowCocoa *) window;
+- (void) resize: (int) width height: (int) height;
 - (void) setClosed;
 - (BOOL) isClosed;
 - (BOOL) canBecomeMainWindow;
@@ -204,15 +205,13 @@ gst_gl_window_cocoa_create_window (GstGLWindowCocoa *window_cocoa)
   NSRect rect = NSMakeRect (0, y, priv->preferred_width, priv->preferred_height);
   NSRect windowRect = NSMakeRect (0, y, priv->preferred_width, priv->preferred_height);
   GstGLContext *context = gst_gl_window_get_context (window);
-  GstGLContextCocoa *context_cocoa;
   GstGLCAOpenGLLayer *layer;
   GstGLNSView *glView;
 
   if (!context)
     return FALSE;
 
-  context_cocoa = GST_GL_CONTEXT_COCOA (context);
-  layer = [[GstGLCAOpenGLLayer alloc] initWithGstGLContext:context];
+  layer = [[GstGLCAOpenGLLayer alloc] initWithGstGLWindow:window];
   layer.autoresizingMask = kCALayerWidthSizable | kCALayerHeightSizable;
   layer.needsDisplayOnBoundsChange = YES;
   glView = [[GstGLNSView alloc] initWithFrameLayer:window_cocoa rect:windowRect layer:layer];
@@ -238,16 +237,14 @@ gst_gl_window_cocoa_create_window (GstGLWindowCocoa *window_cocoa)
     gst_gl_window_cocoa_set_window_handle (window,
         (guintptr) priv->external_view);
 
+  [NSApp setActivationPolicy:NSApplicationActivationPolicyRegular];
+
   return TRUE;
 }
 
 static gboolean
 gst_gl_window_cocoa_open (GstGLWindow *window, GError **err)
 {
-  GstGLWindowCocoa *window_cocoa;
-
-  window_cocoa = GST_GL_WINDOW_COCOA (window);
-
   return TRUE;
 }
 
@@ -329,6 +326,10 @@ _show_window (gpointer data)
   GstGLNSWindow *internal_win_id = (__bridge GstGLNSWindow *)priv->internal_win_id;
 
   GST_DEBUG_OBJECT (window_cocoa, "make the window available\n");
+
+  /* Preferred size might not yet be set when create_window is called,
+   * so we need to resize the window here to be sure. */
+  [internal_win_id resize: priv->preferred_width height: priv->preferred_height];
   [internal_win_id makeMainWindow];
   [internal_win_id orderFrontRegardless];
   [internal_win_id setViewsNeedDisplay:YES];
@@ -634,6 +635,15 @@ gst_gl_window_cocoa_controls_viewport (GstGLWindow * window)
   return self;
 }
 
+- (void) resize: (gint) width height: (gint) height {
+  NSRect frame = [super frame];
+  if (frame.size.width == width && frame.size.height == height)
+    return;
+
+  frame.size = NSMakeSize (width, height);
+  [super setFrame:frame display:YES];
+}
+
 - (void) setClosed {
   m_isClosed = YES;
 }
