@@ -53,8 +53,12 @@
 #include "pbutils.h"
 #include "pbutils-private.h"
 
+/* For g_stat () */
+#include <glib/gstdio.h>
+
 GST_DEBUG_CATEGORY_STATIC (discoverer_debug);
 #define GST_CAT_DEFAULT discoverer_debug
+#define CACHE_DIRNAME "discoverer"
 
 static GQuark _CAPS_QUARK;
 static GQuark _TAGS_QUARK;
@@ -103,6 +107,9 @@ struct _GstDiscovererPrivate
   GError *current_error;
   GstStructure *current_topology;
 
+  GstTagList *all_tags;
+  GstTagList *global_tags;
+
   /* List of private streams */
   GList *streams;
 
@@ -120,12 +127,10 @@ struct _GstDiscovererPrivate
   GstElement *uridecodebin;
   GstBus *bus;
 
-  GType decodebin_type;
-
   /* Custom main context variables */
   GMainContext *ctx;
-  guint sourceid;
-  guint timeoutid;
+  GSource *bus_source;
+  GSource *timeout_source;
 
   /* reusable queries */
   GstQuery *seeking_query;
@@ -135,8 +140,9 @@ struct _GstDiscovererPrivate
   gulong pad_remove_id;
   gulong no_more_pads_id;
   gulong source_chg_id;
-  gulong element_added_id;
   gulong bus_cb_id;
+
+  gboolean use_cache;
 };
 
 #define DISCO_LOCK(dc) g_mutex_lock (&dc->priv->lock);
@@ -158,7 +164,7 @@ _do_init (void)
 };
 
 G_DEFINE_TYPE_EXTENDED (GstDiscoverer, gst_discoverer, G_TYPE_OBJECT, 0,
-    _do_init ());
+    G_ADD_PRIVATE (GstDiscoverer) _do_init ());
 
 enum
 {
@@ -170,11 +176,13 @@ enum
 };
 
 #define DEFAULT_PROP_TIMEOUT 15 * GST_SECOND
+#define DEFAULT_PROP_USE_CACHE FALSE
 
 enum
 {
   PROP_0,
-  PROP_TIMEOUT
+  PROP_TIMEOUT,
+  PROP_USE_CACHE
 };
 
 static guint gst_discoverer_signals[LAST_SIGNAL] = { 0 };
@@ -200,6 +208,13 @@ static void gst_discoverer_set_property (GObject * object, guint prop_id,
     const GValue * value, GParamSpec * pspec);
 static void gst_discoverer_get_property (GObject * object, guint prop_id,
     GValue * value, GParamSpec * pspec);
+static gboolean _setup_locked (GstDiscoverer * dc);
+static void handle_current_async (GstDiscoverer * dc);
+static gboolean emit_discovererd_and_next (GstDiscoverer * dc);
+static GVariant *gst_discoverer_info_to_variant_recurse (GstDiscovererStreamInfo
+    * sinfo, GstDiscovererSerializeFlags flags);
+static GstDiscovererStreamInfo *_parse_discovery (GVariant * variant,
+    GstDiscovererInfo * info);
 
 static void
 gst_discoverer_class_init (GstDiscovererClass * klass)
@@ -212,7 +227,6 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
   gobject_class->set_property = gst_discoverer_set_property;
   gobject_class->get_property = gst_discoverer_get_property;
 
-  g_type_class_add_private (klass, sizeof (GstDiscovererPrivate));
 
   /* properties */
   /**
@@ -229,6 +243,24 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
           GST_SECOND, 3600 * GST_SECOND, DEFAULT_PROP_TIMEOUT,
           G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * GstDiscoverer::use-cache:
+   *
+   * Whether to use a serialized version of the discoverer info from our
+   * own cache if accessible. This allows the discovery to be much faster
+   * as when using this option, we do not need to create a #GstPipeline
+   * and run it, but instead, just reload the #GstDiscovererInfo in its
+   * serialized form.
+   *
+   * The cache files are saved in `$XDG_CACHE_DIR/gstreamer-1.0/discoverer/`.
+   *
+   * Since: 1.16
+   */
+  g_object_class_install_property (gobject_class, PROP_USE_CACHE,
+      g_param_spec_boolean ("use-cache", "use cache", "Use cache",
+          DEFAULT_PROP_USE_CACHE,
+          G_PARAM_READWRITE | G_PARAM_CONSTRUCT | G_PARAM_STATIC_STRINGS));
+
   /* signals */
   /**
    * GstDiscoverer::finished:
@@ -238,8 +270,8 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
    */
   gst_discoverer_signals[SIGNAL_FINISHED] =
       g_signal_new ("finished", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
-      G_STRUCT_OFFSET (GstDiscovererClass, finished),
-      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0, G_TYPE_NONE);
+      G_STRUCT_OFFSET (GstDiscovererClass, finished), NULL, NULL, NULL,
+      G_TYPE_NONE, 0, G_TYPE_NONE);
 
   /**
    * GstDiscoverer::starting:
@@ -249,17 +281,18 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
    */
   gst_discoverer_signals[SIGNAL_STARTING] =
       g_signal_new ("starting", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
-      G_STRUCT_OFFSET (GstDiscovererClass, starting),
-      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0, G_TYPE_NONE);
+      G_STRUCT_OFFSET (GstDiscovererClass, starting), NULL, NULL, NULL,
+      G_TYPE_NONE, 0, G_TYPE_NONE);
 
   /**
    * GstDiscoverer::discovered:
    * @discoverer: the #GstDiscoverer
    * @info: the results #GstDiscovererInfo
-   * @error: (type GLib.Error): #GError, which will be non-NULL if an error
-   *                            occurred during discovery. You must not
-   *                            free this #GError, it will be freed by
-   *                            the discoverer.
+   * @error: (allow-none) (type GLib.Error): #GError, which will be non-NULL
+   *                                         if an error occurred during
+   *                                         discovery. You must not free
+   *                                         this #GError, it will be freed by
+   *                                         the discoverer.
    *
    * Will be emitted in async mode when all information on a URI could be
    * discovered, or an error occurred.
@@ -269,8 +302,7 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
    */
   gst_discoverer_signals[SIGNAL_DISCOVERED] =
       g_signal_new ("discovered", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
-      G_STRUCT_OFFSET (GstDiscovererClass, discovered),
-      NULL, NULL, g_cclosure_marshal_generic,
+      G_STRUCT_OFFSET (GstDiscovererClass, discovered), NULL, NULL, NULL,
       G_TYPE_NONE, 2, GST_TYPE_DISCOVERER_INFO,
       G_TYPE_ERROR | G_SIGNAL_TYPE_STATIC_SCOPE);
 
@@ -290,31 +322,18 @@ gst_discoverer_class_init (GstDiscovererClass * klass)
   gst_discoverer_signals[SIGNAL_SOURCE_SETUP] =
       g_signal_new ("source-setup", G_TYPE_FROM_CLASS (klass),
       G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GstDiscovererClass, source_setup),
-      NULL, NULL, g_cclosure_marshal_generic, G_TYPE_NONE, 1, GST_TYPE_ELEMENT);
-}
-
-static void
-uridecodebin_element_added_cb (GstElement * uridecodebin,
-    GstElement * child, GstDiscoverer * dc)
-{
-  GST_DEBUG ("New element added to uridecodebin : %s",
-      GST_ELEMENT_NAME (child));
-
-  if (G_OBJECT_TYPE (child) == dc->priv->decodebin_type) {
-    g_object_set (child, "post-stream-topology", TRUE, NULL);
-  }
+      NULL, NULL, NULL, G_TYPE_NONE, 1, GST_TYPE_ELEMENT);
 }
 
 static void
 gst_discoverer_init (GstDiscoverer * dc)
 {
-  GstElement *tmp;
   GstFormat format = GST_FORMAT_TIME;
 
-  dc->priv = G_TYPE_INSTANCE_GET_PRIVATE (dc, GST_TYPE_DISCOVERER,
-      GstDiscovererPrivate);
+  dc->priv = gst_discoverer_get_instance_private (dc);
 
   dc->priv->timeout = DEFAULT_PROP_TIMEOUT;
+  dc->priv->use_cache = DEFAULT_PROP_USE_CACHE;
   dc->priv->async = FALSE;
 
   g_mutex_init (&dc->priv->lock);
@@ -325,6 +344,9 @@ gst_discoverer_init (GstDiscoverer * dc)
   dc->priv->target_state = GST_STATE_NULL;
   dc->priv->no_more_pads = FALSE;
 
+  dc->priv->all_tags = NULL;
+  dc->priv->global_tags = NULL;
+
   GST_LOG ("Creating pipeline");
   dc->priv->pipeline = (GstBin *) gst_pipeline_new ("Discoverer");
   GST_LOG_OBJECT (dc, "Creating uridecodebin");
@@ -334,6 +356,9 @@ gst_discoverer_init (GstDiscoverer * dc)
     GST_ERROR ("Can't create uridecodebin");
     return;
   }
+
+  g_object_set (dc->priv->uridecodebin, "post-stream-topology", TRUE, NULL);
+
   GST_LOG_OBJECT (dc, "Adding uridecodebin to pipeline");
   gst_bin_add (dc->priv->pipeline, dc->priv->uridecodebin);
 
@@ -359,16 +384,6 @@ gst_discoverer_init (GstDiscoverer * dc)
 
   GST_DEBUG_OBJECT (dc, "Done initializing Discoverer");
 
-  /* This is ugly. We get the GType of decodebin so we can quickly detect
-   * when a decodebin is added to uridecodebin so we can set the
-   * post-stream-topology setting to TRUE */
-  dc->priv->element_added_id =
-      g_signal_connect_object (dc->priv->uridecodebin, "element-added",
-      G_CALLBACK (uridecodebin_element_added_cb), dc, 0);
-  tmp = gst_element_factory_make ("decodebin", NULL);
-  dc->priv->decodebin_type = G_OBJECT_TYPE (tmp);
-  gst_object_unref (tmp);
-
   /* create queries */
   dc->priv->seeking_query = gst_query_new_seeking (format);
 }
@@ -409,7 +424,6 @@ gst_discoverer_dispose (GObject * obj)
     DISCONNECT_SIGNAL (dc->priv->uridecodebin, dc->priv->pad_remove_id);
     DISCONNECT_SIGNAL (dc->priv->uridecodebin, dc->priv->no_more_pads_id);
     DISCONNECT_SIGNAL (dc->priv->uridecodebin, dc->priv->source_chg_id);
-    DISCONNECT_SIGNAL (dc->priv->uridecodebin, dc->priv->element_added_id);
     DISCONNECT_SIGNAL (dc->priv->bus, dc->priv->bus_cb_id);
 
     /* pipeline was set to NULL in _reset */
@@ -452,6 +466,11 @@ gst_discoverer_set_property (GObject * object, guint prop_id,
     case PROP_TIMEOUT:
       gst_discoverer_set_timeout (dc, g_value_get_uint64 (value));
       break;
+    case PROP_USE_CACHE:
+      DISCO_LOCK (dc);
+      dc->priv->use_cache = g_value_get_boolean (value);
+      DISCO_UNLOCK (dc);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -470,6 +489,11 @@ gst_discoverer_get_property (GObject * object, guint prop_id,
       g_value_set_uint64 (value, dc->priv->timeout);
       DISCO_UNLOCK (dc);
       break;
+    case PROP_USE_CACHE:
+      DISCO_LOCK (dc);
+      g_value_set_boolean (value, dc->priv->use_cache);
+      DISCO_UNLOCK (dc);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -479,6 +503,8 @@ gst_discoverer_get_property (GObject * object, guint prop_id,
 static void
 gst_discoverer_set_timeout (GstDiscoverer * dc, GstClockTime timeout)
 {
+  g_return_if_fail (GST_CLOCK_TIME_IS_VALID (timeout));
+
   GST_DEBUG_OBJECT (dc, "timeout : %" GST_TIME_FORMAT, GST_TIME_ARGS (timeout));
 
   /* FIXME : update current pending timeout if we're running */
@@ -493,28 +519,6 @@ _event_probe (GstPad * pad, GstPadProbeInfo * info, PrivateStream * ps)
   GstEvent *event = GST_PAD_PROBE_INFO_EVENT (info);
 
   switch (GST_EVENT_TYPE (event)) {
-    case GST_EVENT_TAG:{
-      GstTagList *tl = NULL, *tmp;
-
-      gst_event_parse_tag (event, &tl);
-      GST_DEBUG_OBJECT (pad, "tags %" GST_PTR_FORMAT, tl);
-      DISCO_LOCK (ps->dc);
-      /* If preroll is complete, drop these tags - the collected information is
-       * possibly already being processed and adding more tags would be racy */
-      if (G_LIKELY (ps->dc->priv->processing)) {
-        GST_DEBUG_OBJECT (pad, "private stream %p old tags %" GST_PTR_FORMAT,
-            ps, ps->tags);
-        tmp = gst_tag_list_merge (ps->tags, tl, GST_TAG_MERGE_APPEND);
-        if (ps->tags)
-          gst_tag_list_unref (ps->tags);
-        ps->tags = tmp;
-        GST_DEBUG_OBJECT (pad, "private stream %p new tags %" GST_PTR_FORMAT,
-            ps, tmp);
-      } else
-        GST_DEBUG_OBJECT (pad, "Dropping tags since preroll is done");
-      DISCO_UNLOCK (ps->dc);
-      break;
-    }
     case GST_EVENT_TOC:{
       GstToc *tmp;
 
@@ -546,17 +550,30 @@ _event_probe (GstPad * pad, GstPadProbeInfo * info, PrivateStream * ps)
   return GST_PAD_PROBE_OK;
 }
 
-static GstStaticCaps subtitle_caps = GST_STATIC_CAPS ("text/x-raw; "
-    "subpicture/x-pgs; subpicture/x-dvb; subpicture/x-dvd; "
-    "application/x-subtitle-unknown; application/x-ssa; application/x-ass; "
-    "subtitle/x-kate; application/x-kate; subpicture/x-xsub");
+static GstStaticCaps subtitle_caps =
+    GST_STATIC_CAPS
+    ("application/x-ssa; application/x-ass; application/x-kate");
 
 static gboolean
 is_subtitle_caps (const GstCaps * caps)
 {
   GstCaps *subs_caps;
+  GstStructure *s;
+  const gchar *name;
   gboolean ret;
 
+  s = gst_caps_get_structure (caps, 0);
+  if (!s)
+    return FALSE;
+
+  name = gst_structure_get_name (s);
+  if (g_str_has_prefix (name, "text/") ||
+      g_str_has_prefix (name, "subpicture/") ||
+      g_str_has_prefix (name, "subtitle/") ||
+      g_str_has_prefix (name, "closedcaption/") ||
+      g_str_has_prefix (name, "application/x-subtitle"))
+    return TRUE;
+
   subs_caps = gst_static_caps_get (&subtitle_caps);
   ret = gst_caps_can_intersect (caps, subs_caps);
   gst_caps_unref (subs_caps);
@@ -646,13 +663,19 @@ uridecodebin_pad_added_cb (GstElement * uridecodebin, GstPad * pad,
   g_object_set (ps->sink, "silent", TRUE, NULL);
   g_object_set (ps->queue, "max-size-buffers", 1, "silent", TRUE, NULL);
 
-  caps = gst_pad_query_caps (pad, NULL);
-
   sinkpad = gst_element_get_static_pad (ps->queue, "sink");
   if (sinkpad == NULL)
     goto error;
 
-  if (is_subtitle_caps (caps)) {
+  caps = gst_pad_get_current_caps (pad);
+  if (!caps) {
+    GST_WARNING ("Couldn't get negotiated caps from %s:%s",
+        GST_DEBUG_PAD_NAME (pad));
+    caps = gst_pad_query_caps (pad, NULL);
+  }
+
+  if (caps && !gst_caps_is_empty (caps) && !gst_caps_is_any (caps)
+      && is_subtitle_caps (caps)) {
     /* Subtitle streams are sparse and may not provide any information - don't
      * wait for data to preroll */
     ps->probe_id =
@@ -662,7 +685,8 @@ uridecodebin_pad_added_cb (GstElement * uridecodebin, GstPad * pad,
     dc->priv->pending_subtitle_pads++;
   }
 
-  gst_caps_unref (caps);
+  if (caps)
+    gst_caps_unref (caps);
 
   gst_bin_add_many (dc->priv->pipeline, ps->queue, ps->sink, NULL);
 
@@ -1164,7 +1188,7 @@ child_is_raw_stream (const GstCaps * parent, const GstCaps * child)
 }
 
 /* If a parent is non-NULL, collected stream information will be appended to it
- * (and where the information exists, it will be overriden)
+ * (and where the information exists, it will be overridden)
  */
 static GstDiscovererStreamInfo *
 parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
@@ -1192,7 +1216,7 @@ parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
 
     if (nval == NULL) {
       /* FIXME : aggregate with information from main streams */
-      GST_DEBUG ("Coudn't find 'next' ! might be the last entry");
+      GST_DEBUG ("Couldn't find 'next' ! might be the last entry");
     } else {
       GstPad *srcpad;
 
@@ -1236,6 +1260,7 @@ parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
     }
 
     if (add_to_list) {
+      res->stream_number = dc->priv->current_info->stream_count++;
       dc->priv->current_info->stream_list =
           g_list_append (dc->priv->current_info->stream_list, res);
     } else {
@@ -1245,7 +1270,6 @@ parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
   } else if (GST_VALUE_HOLDS_LIST (nval)) {
     guint i, len;
     GstDiscovererContainerInfo *cont;
-    GstTagList *tags;
     GstPad *srcpad;
 
     if (gst_structure_id_get (topology, _ELEMENT_SRCPAD_QUARK, GST_TYPE_PAD,
@@ -1266,26 +1290,10 @@ parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
     cont = (GstDiscovererContainerInfo *)
         g_object_new (GST_TYPE_DISCOVERER_CONTAINER_INFO, NULL);
     cont->parent.caps = caps;
+    if (dc->priv->global_tags)
+      cont->tags = gst_tag_list_ref (dc->priv->global_tags);
     res = (GstDiscovererStreamInfo *) cont;
 
-    if (gst_structure_id_has_field (topology, _TAGS_QUARK)) {
-      GstTagList *tmp;
-
-      gst_structure_id_get (topology, _TAGS_QUARK,
-          GST_TYPE_TAG_LIST, &tags, NULL);
-
-      GST_DEBUG ("Merge tags %" GST_PTR_FORMAT, tags);
-
-      tmp =
-          gst_tag_list_merge (cont->parent.tags, (GstTagList *) tags,
-          GST_TAG_MERGE_APPEND);
-      gst_tag_list_unref (tags);
-      if (cont->parent.tags)
-        gst_tag_list_unref (cont->parent.tags);
-      cont->parent.tags = tmp;
-      GST_DEBUG ("Container info tags %" GST_PTR_FORMAT, tmp);
-    }
-
     for (i = 0; i < len; i++) {
       const GValue *subv = gst_value_list_get_value (nval, i);
       const GstStructure *subst = gst_value_get_structure (subv);
@@ -1305,6 +1313,63 @@ parse_stream_topology (GstDiscoverer * dc, const GstStructure * topology,
   return res;
 }
 
+/* Required DISCO_LOCK to be taken, and will release it */
+static void
+setup_next_uri_locked (GstDiscoverer * dc)
+{
+  if (dc->priv->pending_uris != NULL) {
+    gboolean ready = _setup_locked (dc);
+    DISCO_UNLOCK (dc);
+
+    if (!ready) {
+      /* Start timeout */
+      handle_current_async (dc);
+    } else {
+      g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+          (GSourceFunc) emit_discovererd_and_next, gst_object_ref (dc),
+          gst_object_unref);
+    }
+  } else {
+    /* We're done ! */
+    DISCO_UNLOCK (dc);
+    g_signal_emit (dc, gst_discoverer_signals[SIGNAL_FINISHED], 0);
+  }
+}
+
+static GstDiscovererInfo *
+_ensure_info_tags (GstDiscoverer * dc)
+{
+  GstDiscovererInfo *info = dc->priv->current_info;
+
+  if (dc->priv->all_tags)
+    info->tags = dc->priv->all_tags;
+  dc->priv->all_tags = NULL;
+  return info;
+}
+
+static void
+emit_discovererd (GstDiscoverer * dc)
+{
+  GstDiscovererInfo *info = _ensure_info_tags (dc);
+  GST_DEBUG_OBJECT (dc, "Emitting 'discoverered' %s", info->uri);
+  g_signal_emit (dc, gst_discoverer_signals[SIGNAL_DISCOVERED], 0,
+      info, dc->priv->current_error);
+  /* Clients get a copy of current_info since it is a boxed type */
+  gst_discoverer_info_unref (dc->priv->current_info);
+  dc->priv->current_info = NULL;
+}
+
+static gboolean
+emit_discovererd_and_next (GstDiscoverer * dc)
+{
+  emit_discovererd (dc);
+
+  DISCO_LOCK (dc);
+  setup_next_uri_locked (dc);
+
+  return G_SOURCE_REMOVE;
+}
+
 /* Called when pipeline is pre-rolled */
 static void
 discoverer_collect (GstDiscoverer * dc)
@@ -1312,9 +1377,17 @@ discoverer_collect (GstDiscoverer * dc)
   GST_DEBUG ("Collecting information");
 
   /* Stop the timeout handler if present */
-  if (dc->priv->timeoutid) {
-    g_source_remove (dc->priv->timeoutid);
-    dc->priv->timeoutid = 0;
+  if (dc->priv->timeout_source) {
+    g_source_destroy (dc->priv->timeout_source);
+    g_source_unref (dc->priv->timeout_source);
+    dc->priv->timeout_source = NULL;
+  }
+
+  if (dc->priv->use_cache && dc->priv->current_info
+      && dc->priv->current_info->from_cache) {
+    GST_DEBUG_OBJECT (dc,
+        "Nothing to collect as the info was built from" " the cache");
+    return;
   }
 
   if (dc->priv->streams) {
@@ -1380,9 +1453,13 @@ discoverer_collect (GstDiscoverer * dc)
     else
       dc->priv->current_info->live = TRUE;
 
-    if (dc->priv->current_topology)
+    if (dc->priv->current_topology) {
+      dc->priv->current_info->stream_count = 1;
       dc->priv->current_info->stream_info = parse_stream_topology (dc,
           dc->priv->current_topology, NULL);
+      if (dc->priv->current_info->stream_info)
+        dc->priv->current_info->stream_info->stream_number = 0;
+    }
 
     /*
      * Images need some special handling. They do not have a duration, have
@@ -1408,14 +1485,18 @@ discoverer_collect (GstDiscoverer * dc)
     }
   }
 
-  if (dc->priv->async) {
-    GST_DEBUG ("Emitting 'discoverered'");
-    g_signal_emit (dc, gst_discoverer_signals[SIGNAL_DISCOVERED], 0,
-        dc->priv->current_info, dc->priv->current_error);
-    /* Clients get a copy of current_info since it is a boxed type */
-    gst_discoverer_info_unref (dc->priv->current_info);
-    dc->priv->current_info = NULL;
+  if (dc->priv->use_cache && dc->priv->current_info->cachefile &&
+      dc->priv->current_info->result == GST_DISCOVERER_OK) {
+    GVariant *variant = gst_discoverer_info_to_variant (dc->priv->current_info,
+        GST_DISCOVERER_SERIALIZE_ALL);
+
+    g_file_set_contents (dc->priv->current_info->cachefile,
+        g_variant_get_data (variant), g_variant_get_size (variant), NULL);
+    g_variant_unref (variant);
   }
+
+  if (dc->priv->async)
+    emit_discovererd (dc);
 }
 
 static void
@@ -1446,8 +1527,8 @@ handle_current_async (GstDiscoverer * dc)
   /* Attach a timeout to the main context */
   source = g_timeout_source_new (dc->priv->timeout / GST_MSECOND);
   g_source_set_callback_indirect (source, g_object_ref (dc), &cb_funcs);
-  dc->priv->timeoutid = g_source_attach (source, dc->priv->ctx);
-  g_source_unref (source);
+  g_source_attach (source, dc->priv->ctx);
+  dc->priv->timeout_source = source;
 }
 
 
@@ -1580,23 +1661,41 @@ handle_message (GstDiscoverer * dc, GstMessage * msg)
 
     case GST_MESSAGE_TAG:
     {
-      GstTagList *tl, *tmp;
+      GstTagList *tl, *tmp = NULL;
+      GstTagScope scope;
 
       gst_message_parse_tag (msg, &tl);
+      scope = gst_tag_list_get_scope (tl);
       GST_DEBUG_OBJECT (GST_MESSAGE_SRC (msg), "Got tags %" GST_PTR_FORMAT, tl);
-      /* Merge with current tags */
-      tmp =
-          gst_tag_list_merge (dc->priv->current_info->tags, tl,
-          GST_TAG_MERGE_APPEND);
+
+      tmp = gst_tag_list_merge (dc->priv->all_tags, tl, GST_TAG_MERGE_APPEND);
+      if (dc->priv->all_tags)
+        gst_tag_list_unref (dc->priv->all_tags);
+      dc->priv->all_tags = tmp;
+
+      if (scope == GST_TAG_SCOPE_STREAM) {
+        for (GList * curr = dc->priv->streams; curr; curr = curr->next) {
+          PrivateStream *ps = (PrivateStream *) curr->data;
+          if (GST_MESSAGE_SRC (msg) == GST_OBJECT_CAST (ps->sink)) {
+            tmp = gst_tag_list_merge (ps->tags, tl, GST_TAG_MERGE_APPEND);
+            if (ps->tags)
+              gst_tag_list_unref (ps->tags);
+            ps->tags = tmp;
+            GST_DEBUG_OBJECT (ps->pad, "tags %" GST_PTR_FORMAT, ps->tags);
+            break;
+          }
+        }
+      } else {
+        tmp =
+            gst_tag_list_merge (dc->priv->global_tags, tl,
+            GST_TAG_MERGE_APPEND);
+        if (dc->priv->global_tags)
+          gst_tag_list_unref (dc->priv->global_tags);
+        dc->priv->global_tags = tmp;
+      }
       gst_tag_list_unref (tl);
-      if (dc->priv->current_info->tags)
-        gst_tag_list_unref (dc->priv->current_info->tags);
-      dc->priv->current_info->tags = tmp;
-      GST_DEBUG_OBJECT (GST_MESSAGE_SRC (msg), "Current info %p, tags %"
-          GST_PTR_FORMAT, dc->priv->current_info, tmp);
     }
       break;
-
     case GST_MESSAGE_TOC:
     {
       GstToc *tmp;
@@ -1660,19 +1759,118 @@ handle_current_sync (GstDiscoverer * dc)
   g_timer_destroy (timer);
 }
 
-static void
+static gchar *
+_serialized_info_get_path (GstDiscoverer * dc, gchar * uri)
+{
+  GChecksum *cs = NULL;
+  GStatBuf file_status;
+  gchar *location = NULL, *res = NULL, *cache_dir = NULL, *tmp = NULL,
+      *protocol = gst_uri_get_protocol (uri), hash_dirname[3] = "00";
+  const gchar *checksum;
+
+  if (g_ascii_strcasecmp (protocol, "file") != 0) {
+    GST_DEBUG_OBJECT (dc, "Can not work with serialized DiscovererInfo"
+        " on non local files - protocol: %s", protocol);
+
+    goto done;
+  }
+
+  location = gst_uri_get_location (uri);
+  if (g_stat (location, &file_status) < 0) {
+    GST_DEBUG_OBJECT (dc, "Could not get stat for file: %s", location);
+
+    goto done;
+  }
+
+  tmp = g_strdup_printf ("%s-%" G_GSIZE_FORMAT "-%" G_GINT64_FORMAT,
+      location, (gsize) file_status.st_size, (gint64) file_status.st_mtime);
+  cs = g_checksum_new (G_CHECKSUM_SHA1);
+  g_checksum_update (cs, (const guchar *) tmp, strlen (tmp));
+  checksum = g_checksum_get_string (cs);
+
+  hash_dirname[0] = checksum[0];
+  hash_dirname[1] = checksum[1];
+  cache_dir =
+      g_build_filename (g_get_user_cache_dir (), "gstreamer-" GST_API_VERSION,
+      CACHE_DIRNAME, hash_dirname, NULL);
+  g_mkdir_with_parents (cache_dir, 0777);
+
+  res = g_build_filename (cache_dir, &checksum[2], NULL);
+
+done:
+  g_checksum_free (cs);
+  g_free (cache_dir);
+  g_free (location);
+  g_free (tmp);
+  g_free (protocol);
+
+  return res;
+}
+
+static GstDiscovererInfo *
+_get_info_from_cachefile (GstDiscoverer * dc, gchar * cachefile)
+{
+  gchar *data;
+  gsize length;
+
+  if (g_file_get_contents (cachefile, &data, &length, NULL)) {
+    GstDiscovererInfo *info = NULL;
+    GVariant *variant =
+        g_variant_new_from_data (G_VARIANT_TYPE ("v"), data, length,
+        TRUE, NULL, NULL);
+
+    info = gst_discoverer_info_from_variant (variant);
+    g_variant_unref (variant);
+
+    if (info) {
+      info->cachefile = cachefile;
+      info->from_cache = (gpointer) 0x01;
+    }
+
+    GST_INFO_OBJECT (dc, "Got info from cache: %p", info);
+    g_free (data);
+
+    return info;
+  }
+
+  return NULL;
+}
+
+static gboolean
 _setup_locked (GstDiscoverer * dc)
 {
   GstStateChangeReturn ret;
+  gchar *uri = (gchar *) dc->priv->pending_uris->data;
+  gchar *cachefile = NULL;
+
+  dc->priv->pending_uris =
+      g_list_delete_link (dc->priv->pending_uris, dc->priv->pending_uris);
+
+  if (dc->priv->use_cache) {
+    cachefile = _serialized_info_get_path (dc, uri);
+    if (cachefile)
+      dc->priv->current_info = _get_info_from_cachefile (dc, cachefile);
+
+    if (dc->priv->current_info) {
+      /* Make sure the URI is exactly what the user passed in */
+      g_free (dc->priv->current_info->uri);
+      dc->priv->current_info->uri = uri;
+
+      dc->priv->current_info->cachefile = cachefile;
+      dc->priv->processing = FALSE;
+      dc->priv->target_state = GST_STATE_NULL;
+
+      return TRUE;
+    }
+  }
 
   GST_DEBUG ("Setting up");
 
   /* Pop URI off the pending URI list */
   dc->priv->current_info =
       (GstDiscovererInfo *) g_object_new (GST_TYPE_DISCOVERER_INFO, NULL);
-  dc->priv->current_info->uri = (gchar *) dc->priv->pending_uris->data;
-  dc->priv->pending_uris =
-      g_list_delete_link (dc->priv->pending_uris, dc->priv->pending_uris);
+  dc->priv->current_info->cachefile = cachefile;
+  dc->priv->current_info->uri = uri;
 
   /* set uri on uridecodebin */
   g_object_set (dc->priv->uridecodebin, "uri", dc->priv->current_info->uri,
@@ -1703,6 +1901,8 @@ _setup_locked (GstDiscoverer * dc)
 
   GST_DEBUG_OBJECT (dc, "Pipeline going to PAUSED : %s",
       gst_element_state_change_return_get_name (ret));
+
+  return FALSE;
 }
 
 static void
@@ -1737,6 +1937,16 @@ discoverer_cleanup (GstDiscoverer * dc)
 
   dc->priv->current_info = NULL;
 
+  if (dc->priv->all_tags) {
+    gst_tag_list_unref (dc->priv->all_tags);
+    dc->priv->all_tags = NULL;
+  }
+
+  if (dc->priv->global_tags) {
+    gst_tag_list_unref (dc->priv->global_tags);
+    dc->priv->global_tags = NULL;
+  }
+
   dc->priv->pending_subtitle_pads = 0;
 
   dc->priv->current_state = GST_STATE_NULL;
@@ -1747,16 +1957,7 @@ discoverer_cleanup (GstDiscoverer * dc)
 
   /* Try popping the next uri */
   if (dc->priv->async) {
-    if (dc->priv->pending_uris != NULL) {
-      _setup_locked (dc);
-      DISCO_UNLOCK (dc);
-      /* Start timeout */
-      handle_current_async (dc);
-    } else {
-      /* We're done ! */
-      DISCO_UNLOCK (dc);
-      g_signal_emit (dc, gst_discoverer_signals[SIGNAL_FINISHED], 0);
-    }
+    setup_next_uri_locked (dc);
   } else
     DISCO_UNLOCK (dc);
 
@@ -1783,7 +1984,6 @@ static gboolean
 async_timeout_cb (GstDiscoverer * dc)
 {
   if (!g_source_is_destroyed (g_main_current_source ())) {
-    dc->priv->timeoutid = 0;
     GST_DEBUG ("Setting result to TIMEOUT");
     dc->priv->current_info->result = GST_DISCOVERER_TIMEOUT;
     dc->priv->processing = FALSE;
@@ -1802,6 +2002,7 @@ async_timeout_cb (GstDiscoverer * dc)
 static GstDiscovererResult
 start_discovering (GstDiscoverer * dc)
 {
+  gboolean ready;
   GstDiscovererResult res = GST_DISCOVERER_OK;
 
   GST_DEBUG ("Starting");
@@ -1823,14 +2024,27 @@ start_discovering (GstDiscoverer * dc)
 
   g_signal_emit (dc, gst_discoverer_signals[SIGNAL_STARTING], 0);
 
-  _setup_locked (dc);
+  ready = _setup_locked (dc);
 
   DISCO_UNLOCK (dc);
 
-  if (dc->priv->async)
+  if (dc->priv->async) {
+    if (ready) {
+      GSource *source;
+
+      source = g_idle_source_new ();
+      g_source_set_callback (source,
+          (GSourceFunc) emit_discovererd_and_next, gst_object_ref (dc),
+          gst_object_unref);
+      g_source_attach (source, dc->priv->ctx);
+      goto beach;
+    }
+
     handle_current_async (dc);
-  else
-    handle_current_sync (dc);
+  } else {
+    if (!ready)
+      handle_current_sync (dc);
+  }
 
 beach:
   return res;
@@ -1843,6 +2057,7 @@ _serialize_common_stream_info (GstDiscovererStreamInfo * sinfo,
     GstDiscovererSerializeFlags flags)
 {
   GVariant *common;
+  GVariant *nextv = NULL;
   gchar *caps_str = NULL, *tags_str = NULL, *misc_str = NULL;
 
   if (sinfo->caps && (flags & GST_DISCOVERER_SERIALIZE_CAPS))
@@ -1854,9 +2069,15 @@ _serialize_common_stream_info (GstDiscovererStreamInfo * sinfo,
   if (sinfo->misc && (flags & GST_DISCOVERER_SERIALIZE_MISC))
     misc_str = gst_structure_to_string (sinfo->misc);
 
+
+  if (sinfo->next)
+    nextv = gst_discoverer_info_to_variant_recurse (sinfo->next, flags);
+  else
+    nextv = g_variant_new ("()");
+
   common =
-      g_variant_new ("(msmsmsms)", sinfo->stream_id, caps_str, tags_str,
-      misc_str);
+      g_variant_new ("(msmsmsmsv)", sinfo->stream_id, caps_str, tags_str,
+      misc_str, nextv);
 
   g_free (caps_str);
   g_free (tags_str);
@@ -1961,6 +2182,16 @@ gst_discoverer_info_to_variant_recurse (GstDiscovererStreamInfo * sinfo,
     stream_variant =
         g_variant_new ("(yvv)", 's', common_stream_variant,
         subtitle_stream_info);
+  } else {
+    GVariant *nextv = NULL;
+    GstDiscovererStreamInfo *ninfo =
+        gst_discoverer_stream_info_get_next (sinfo);
+
+    nextv = gst_discoverer_info_to_variant_recurse (ninfo, flags);
+
+    stream_variant =
+        g_variant_new ("(yvv)", 'n', common_stream_variant,
+        g_variant_new ("v", nextv));
   }
 
   return stream_variant;
@@ -2008,7 +2239,8 @@ _parse_info (GstDiscovererInfo * info, GVariant * info_variant)
 }
 
 static void
-_parse_common_stream_info (GstDiscovererStreamInfo * sinfo, GVariant * common)
+_parse_common_stream_info (GstDiscovererStreamInfo * sinfo, GVariant * common,
+    GstDiscovererInfo * info)
 {
   const gchar *str;
 
@@ -2028,6 +2260,16 @@ _parse_common_stream_info (GstDiscovererStreamInfo * sinfo, GVariant * common)
   if (str)
     sinfo->misc = gst_structure_new_from_string (str);
 
+  if (g_variant_n_children (common) > 4) {
+    GVariant *nextv;
+
+    GET_FROM_TUPLE (common, variant, 4, &nextv);
+    if (g_variant_n_children (nextv) > 0) {
+      sinfo->next = _parse_discovery (nextv, info);
+    }
+    g_variant_unref (nextv);
+  }
+
   g_variant_unref (common);
 }
 
@@ -2113,14 +2355,20 @@ _parse_discovery (GVariant * variant, GstDiscovererInfo * info)
       _parse_subtitle_stream_info (GST_DISCOVERER_SUBTITLE_INFO (sinfo),
           g_variant_get_child_value (specific, 0));
       break;
+    case 'n':
+      sinfo = g_object_new (GST_TYPE_DISCOVERER_STREAM_INFO, NULL);
+      break;
     default:
       GST_WARNING ("Unexpected discoverer info type %d", type);
       goto out;
   }
 
-  _parse_common_stream_info (sinfo, g_variant_get_child_value (common, 0));
+  _parse_common_stream_info (sinfo, g_variant_get_child_value (common, 0),
+      info);
 
-  info->stream_list = g_list_append (info->stream_list, sinfo);
+  if (!GST_IS_DISCOVERER_CONTAINER_INFO (sinfo)) {
+    info->stream_list = g_list_append (info->stream_list, sinfo);
+  }
 
   if (!info->stream_info) {
     info->stream_info = sinfo;
@@ -2132,6 +2380,7 @@ _parse_discovery (GVariant * variant, GstDiscovererInfo * info)
 
     GstDiscovererContainerInfo *cinfo = GST_DISCOVERER_CONTAINER_INFO (sinfo);
     g_variant_iter_init (&iter, specific);
+    cinfo->tags = sinfo->tags;
     while ((child = g_variant_iter_next_value (&iter))) {
       GstDiscovererStreamInfo *child_info;
       child_info = _parse_discovery (g_variant_get_variant (child), info);
@@ -2187,8 +2436,8 @@ gst_discoverer_start (GstDiscoverer * discoverer)
   source = gst_bus_create_watch (discoverer->priv->bus);
   g_source_set_callback (source, (GSourceFunc) gst_bus_async_signal_func,
       NULL, NULL);
-  discoverer->priv->sourceid = g_source_attach (source, ctx);
-  g_source_unref (source);
+  g_source_attach (source, ctx);
+  discoverer->priv->bus_source = source;
   discoverer->priv->ctx = g_main_context_ref (ctx);
 
   start_discovering (discoverer);
@@ -2230,14 +2479,16 @@ gst_discoverer_stop (GstDiscoverer * discoverer)
   DISCO_UNLOCK (discoverer);
 
   /* Remove timeout handler */
-  if (discoverer->priv->timeoutid) {
-    g_source_remove (discoverer->priv->timeoutid);
-    discoverer->priv->timeoutid = 0;
+  if (discoverer->priv->timeout_source) {
+    g_source_destroy (discoverer->priv->timeout_source);
+    g_source_unref (discoverer->priv->timeout_source);
+    discoverer->priv->timeout_source = NULL;
   }
   /* Remove signal watch */
-  if (discoverer->priv->sourceid) {
-    g_source_remove (discoverer->priv->sourceid);
-    discoverer->priv->sourceid = 0;
+  if (discoverer->priv->bus_source) {
+    g_source_destroy (discoverer->priv->bus_source);
+    g_source_unref (discoverer->priv->bus_source);
+    discoverer->priv->bus_source = NULL;
   }
   /* Unref main context */
   if (discoverer->priv->ctx) {
@@ -2294,7 +2545,7 @@ gst_discoverer_discover_uri_async (GstDiscoverer * discoverer,
  * gst_discoverer_discover_uri:
  * @discoverer: A #GstDiscoverer
  * @uri: The URI to run on.
- * @err: (out) (allow-none): If an error occurred, this field will be filled in.
+ * @err: If an error occurred, this field will be filled in.
  *
  * Synchronously discovers the given @uri.
  *
@@ -2345,7 +2596,7 @@ gst_discoverer_discover_uri (GstDiscoverer * discoverer, const gchar * uri,
         discoverer->priv->current_info->result);
     discoverer->priv->current_info->result = res;
   }
-  info = discoverer->priv->current_info;
+  info = _ensure_info_tags (discoverer);
 
   discoverer_cleanup (discoverer);
 
@@ -2370,6 +2621,8 @@ gst_discoverer_new (GstClockTime timeout, GError ** err)
 {
   GstDiscoverer *res;
 
+  g_return_val_if_fail (GST_CLOCK_TIME_IS_VALID (timeout), NULL);
+
   res = g_object_new (GST_TYPE_DISCOVERER, "timeout", timeout, NULL);
   if (res->priv->uridecodebin == NULL) {
     if (err)
@@ -2403,7 +2656,7 @@ gst_discoverer_info_to_variant (GstDiscovererInfo * info,
 {
   /* FIXME: implement TOC support */
   GVariant *stream_variant;
-  GVariant *variant;
+  GVariant *variant, *info_variant;
   GstDiscovererStreamInfo *sinfo;
   GVariant *wrapper;
 
@@ -2413,10 +2666,11 @@ gst_discoverer_info_to_variant (GstDiscovererInfo * info,
 
   sinfo = gst_discoverer_info_get_stream_info (info);
   stream_variant = gst_discoverer_info_to_variant_recurse (sinfo, flags);
-  variant =
-      g_variant_new ("(vv)", _serialize_info (info, flags), stream_variant);
+  info_variant = _serialize_info (info, flags);
 
-  /* Returning a wrapper implies some small overhead, but simplifies 
+  variant = g_variant_new ("(vv)", info_variant, stream_variant);
+
+  /* Returning a wrapper implies some small overhead, but simplifies
    * deserializing from bytes */
   wrapper = g_variant_new_variant (variant);
 
