@@ -39,10 +39,8 @@
 #include "config.h"
 #endif
 
+#include "gstaudiomixerelements.h"
 #include "gstaudiointerleave.h"
-#include <gst/audio/audio.h>
-
-#include <string.h>
 
 #define GST_CAT_DEFAULT gst_audio_interleave_debug
 GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
@@ -55,6 +53,8 @@ enum
 
 G_DEFINE_TYPE (GstAudioInterleavePad, gst_audio_interleave_pad,
     GST_TYPE_AUDIO_AGGREGATOR_PAD);
+GST_ELEMENT_REGISTER_DEFINE_WITH_CODE (audiointerleave, "audiointerleave",
+    GST_RANK_NONE, GST_TYPE_AUDIO_INTERLEAVE, audiomixer_element_init (plugin));
 
 static void
 gst_audio_interleave_pad_get_property (GObject * object, guint prop_id,
@@ -535,12 +535,12 @@ static gboolean
 gst_audio_interleave_negotiated_src_caps (GstAggregator * agg, GstCaps * caps)
 {
   GstAudioInterleave *self = GST_AUDIO_INTERLEAVE (agg);
-  GstAudioAggregator *aagg = GST_AUDIO_AGGREGATOR (self);
+  GstAudioAggregatorPad *srcpad = GST_AUDIO_AGGREGATOR_PAD (agg->srcpad);
 
   if (!GST_AGGREGATOR_CLASS (parent_class)->negotiated_src_caps (agg, caps))
     return FALSE;
 
-  gst_audio_interleave_set_process_function (self, &aagg->info);
+  gst_audio_interleave_set_process_function (self, &srcpad->info);
 
   return TRUE;
 }
@@ -560,8 +560,8 @@ gst_audio_interleave_class_init (GstAudioInterleaveClass * klass)
   gobject_class->get_property = gst_audio_interleave_get_property;
   gobject_class->finalize = gst_audio_interleave_finalize;
 
-  gst_element_class_add_static_pad_template (gstelement_class,
-      &gst_audio_interleave_src_template);
+  gst_element_class_add_static_pad_template_with_gtype (gstelement_class,
+      &gst_audio_interleave_src_template, GST_TYPE_AUDIO_AGGREGATOR_PAD);
   gst_element_class_add_static_pad_template_with_gtype (gstelement_class,
       &gst_audio_interleave_sink_template, GST_TYPE_AUDIO_INTERLEAVE_PAD);
   gst_element_class_set_static_metadata (gstelement_class, "AudioInterleave",
@@ -580,10 +580,9 @@ gst_audio_interleave_class_init (GstAudioInterleaveClass * klass)
   agg_class->negotiated_src_caps = gst_audio_interleave_negotiated_src_caps;
 
   aagg_class->aggregate_one_buffer = gst_audio_interleave_aggregate_one_buffer;
-  aagg_class->convert_buffer = NULL;
 
   /**
-   * GstInterleave:channel-positions
+   * GstAudioInterleave:channel-positions:
    *
    * Channel positions: This property controls the channel positions
    * that are used on the src caps. The number of elements should be
@@ -607,7 +606,7 @@ gst_audio_interleave_class_init (GstAudioInterleaveClass * klass)
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
   /**
-   * GstInterleave:channel-positions-from-input
+   * GstAudioInterleave:channel-positions-from-input:
    *
    * Channel positions from input: If this property is set to %TRUE the channel
    * positions will be taken from the input caps if valid channel positions for
@@ -621,6 +620,8 @@ gst_audio_interleave_class_init (GstAudioInterleaveClass * klass)
           "Channel positions from input",
           "Take channel positions from the input", TRUE,
           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  gst_type_mark_as_plugin_api (GST_TYPE_AUDIO_INTERLEAVE_PAD, 0);
 }
 
 static void
@@ -819,14 +820,16 @@ gst_audio_interleave_aggregate_one_buffer (GstAudioAggregator * aagg,
   GstMapInfo outmap;
   gint out_width, in_bpf, out_bpf, out_channels, channel;
   guint8 *outdata;
+  GstAggregator *agg = GST_AGGREGATOR (aagg);
+  GstAudioAggregatorPad *srcpad = GST_AUDIO_AGGREGATOR_PAD (agg->srcpad);
 
   GST_OBJECT_LOCK (aagg);
   GST_OBJECT_LOCK (aaggpad);
 
-  out_width = GST_AUDIO_INFO_WIDTH (&aagg->info) / 8;
+  out_width = GST_AUDIO_INFO_WIDTH (&srcpad->info) / 8;
   in_bpf = GST_AUDIO_INFO_BPF (&aaggpad->info);
-  out_bpf = GST_AUDIO_INFO_BPF (&aagg->info);
-  out_channels = GST_AUDIO_INFO_CHANNELS (&aagg->info);
+  out_bpf = GST_AUDIO_INFO_BPF (&srcpad->info);
+  out_channels = GST_AUDIO_INFO_CHANNELS (&srcpad->info);
 
   gst_buffer_map (outbuf, &outmap, GST_MAP_READWRITE);
   gst_buffer_map (inbuf, &inmap, GST_MAP_READ);
@@ -894,7 +897,7 @@ gst_audio_interleave_child_proxy_init (gpointer g_iface, gpointer iface_data)
 {
   GstChildProxyInterface *iface = g_iface;
 
-  GST_INFO ("intializing child proxy interface");
+  GST_INFO ("initializing child proxy interface");
   iface->get_child_by_index =
       gst_audio_interleave_child_proxy_get_child_by_index;
   iface->get_children_count =
