@@ -29,7 +29,7 @@
 #include <string.h>
 #include <gst/gst.h>
 
-#include <gst/gst-i18n-plugin.h>
+#include <glib/gi18n-lib.h>
 #include <gst/pbutils/pbutils.h>
 
 #include "gstplaybackelements.h"
